from pathlib import Path
from typing import Dict, List, Optional, Union
import yaml

import pandas as pd

from opa_tools.config import get_request_dir
from .transformation import flatten
from .utils import (
    ALLOWED_KINDS,
    normalize_request_identifier,
    format_identifier,
    get_request_location,
)


REQUEST_DIR = get_request_dir()


def _yaml_read(path: Path):
    with open(path, "r", encoding="utf-8") as rf:
        return yaml.safe_load(rf)

def _get_single_request(
        name: str,
        subrequest: Union[str, None],
        month: str,
        kind: str,
        allow_subrequests: bool,
        no_exist_ok: bool
    ) -> Dict:
    return _yaml_read(
        get_request_location(
            name,
            subrequest=subrequest,
            month=month,
            kind=kind,
            allow_subrequests=allow_subrequests,
            no_exist_ok=False,
        )
    )

def _subdir_iterator(path: Path, level=0, only_terminal=False):
    if level not in {0, 1}:
        raise NotImplementedError(
            "Level must be either 0 or 1."
        )
    for subdir in sorted(path.iterdir(), key=lambda p: p.name):
        if not subdir.is_dir():
            continue
        if not only_terminal:
            yield subdir
            continue
        is_terminal_subdir = not any(
            subsubdir.is_dir() for subsubdir in subdir.iterdir()
        )
        if is_terminal_subdir:
            yield subdir
            continue
        elif level > 0:
            yield from _subdir_iterator(subdir, level - 1, only_terminal=True)
            continue

def get_request_identifiers(month: str = "*", subrequests: bool = False):
    if month == "*":
        output = []
        for month in sorted(path.name for path in REQUEST_DIR.iterdir()):
            output.extend(get_request_identifiers(month=month, subrequests=subrequests))
        return output
    month_dir = REQUEST_DIR / month
    if not month_dir.exists():
        raise FileNotFoundError(
            f"Month directory {month_dir} does not exists."
        )
    output = []
    for subdir in _subdir_iterator(path=month_dir, level=1, only_terminal=subrequests):
        subdir_parts = subdir.relative_to(REQUEST_DIR).parts
        # TODO: Check that is really a month
        output.append(format_identifier(*subdir_parts))
    return output


def _get_requests(
    name: str,
    subrequest: Union[str, None],
    month: str,
    kind: str,
    flat: bool,
    allow_subrequests: bool,
) -> List[Dict] | Dict:
    if kind == "*":
        no_exist_error = ("gsv", "opa")
        requests = {
            kind: _get_single_request(
                name=name,
                subrequest=subrequest,
                month=month,
                kind=kind,
                allow_subrequests=allow_subrequests,
                no_exist_ok=(kind not in no_exist_error),
            ) for kind in ALLOWED_KINDS
        }
        if flat:
            return flatten(**requests)
        return [requests[kind] for kind in ALLOWED_KINDS]
    return [
        _get_single_request(
            name=name,
            subrequest=subrequest,
            month=month,
            kind=kind,
            allow_subrequests=allow_subrequests,
            no_exist_ok=False,
        )
    ]

def get_request(
    identifier: Optional[Union[str, int, float]] = None,
    /,
    subrequest: Optional[str] = None,
    month: Optional[str] = None,
    kind: str = "*",
    *,
    allow_subrequests=True,
    flat=False,
    single_as_list=False,
) -> Union[List[Dict], Dict]:
    if identifier is None:
        if month is None:
            month = "*"
        output = []
        for identifier in get_request_identifiers(month=month):
            output.append(
                get_request(identifier, kind=kind, flat=flat, single_as_list=True)
            )
        return output
    name, subrequest, month, kind = normalize_request_identifier(
        identifier, subrequest=subrequest, month=month, kind=kind
    )
    requests = _get_requests(
        name=name,
        subrequest=subrequest,
        month=month,
        kind=kind,
        flat=flat,
        allow_subrequests=allow_subrequests,
    )
    if len(requests) == 1 and not single_as_list:
        return requests[0]
    return requests
