import json
import re
import pytest
from pathlib import Path

import opa_tools.config
from opa_tools.request.retrieval import get_request, get_request_identifiers
from opa_tools.request.utils import ALLOWED_KINDS


@pytest.fixture(scope="session")
def monkeysession():
    with pytest.MonkeyPatch.context() as mp:
        yield mp

@pytest.fixture(scope="session")
def request_dir(tmpdir_factory, monkeysession):
    tmpdir = tmpdir_factory.mktemp("data")
    monkeysession.setattr(opa_tools.config, "get_request_dir", tmpdir)
    with open("requests_testdata.json", "r") as rf:
        json_requests = json.load(rf)
    for relpath, request in json_requests.items():
        Path(tmpdir / relpath).parent.mkdir(parents=True, exist_ok=True)
        with open(tmpdir / relpath, "w") as wf:
            wf.write(request)
    return Path(tmpdir)


def test_get_request_no_flat(request_dir):
    request = get_request(1, kind="*", flat=False, single_as_list=False)
    assert isinstance(request, list)
    assert len(request) > 0
    assert len(request) == len(ALLOWED_KINDS)


def test_get_request_flat(request_dir):
    request = get_request(1, kind="*", flat=True, single_as_list=False)
    assert isinstance(request, dict)
    assert len(request) > 0
    assert all(
        any(
            key.startswith(kind)
            for kind
            in ALLOWED_KINDS
        ) for key in request
    )

@pytest.mark.parametrize("kind", ["gsv", "opa"])
def test_get_request_kind(request_dir, kind):
    request_gsv = get_request(1, kind=kind, flat=True, single_as_list=False)
    request = get_request(1, flat=True, single_as_list=False)
    assert isinstance(request, dict) and isinstance(request_gsv, dict)
    assert len(request) > 0 and len(request_gsv) > 0
    filtered_request = {
        str.removeprefix(key, f"{kind}_"): value
        for key, value
        in request.items()
        if key.startswith(f"{kind}_")
    }
    assert filtered_request == request_gsv

def test_get_identifiers(request_dir):
    all_identifiers = []
    for month in request_dir.iterdir():
        req_names = [path.name for path in month.iterdir()]
        identifiers = [f"{month.name}/{req_name}" for req_name in req_names]
        assert set(identifiers) == set(get_request_identifiers(month))
        all_identifiers.extend(identifiers)
    assert set(all_identifiers) == set(get_request_identifiers())

